///////////////////////////////////////////////////////////////////
// CExecEnv.h
// 
// Execution Environment
// Encapsulates execution parameters, attributes, and return values
///////////////////////////////////////////////////////////////////

#ifndef EXECENVDefined
#define EXECENVDefined 1



////////////////////////////////////////////////////////////////
// Structure: ATTRIBUTE
// Each node holds one attribute in an attribute table.  The 
// Attribute Table is used for storing and manipulating:
//   * Declared Genetic Attributes (Variables)
//   * User-Function Arguments
//   * Input-Output base program arguments
//   * Function Return variables
// Used by class CAttributeTable as a linked list.
struct ATTRIBUTE
{
	CString Name;
	enum RETURNTYPE Type;
	CString Value;
	int Level;
	struct ATTRIBUTE *next;
	struct ATTRIBUTE *prev;
};

////////////////////////////////////////////////////////////////
// Structure: FUNCTIONNODE
// Each node holds one user-created function; is used by class 
// CFunctionTable as a linked list.
struct FUNCTIONNODE
{
	CString Name;
	enum RETURNTYPE Type;
	class CGPProgram * ExecTree;
	struct FUNCTIONNODE *next;
	int NumParameters;
	CString Parameters[25];  // Hopefully and assumedly, no more parameters than that!
	enum RETURNTYPE ParameterTypes[25];
};

struct FUNCSELECTIONNODE
{
	CString Name;
 	bool UserFunction;       // TRUE for user-function; FALSE for selected built-in
	struct FUNCSELECTIONNODE *next;
};

////////////////////////////////////////////////////////////////
// Class CAttributeTable
// Encapsulates attributes (See Struct ATTRIBUTE) and methods to
// declare, manipulate, and return them.  
class CAttributeTable
{
public:
	CAttributeTable();
	~CAttributeTable();

	bool AddAttribute (RETURNTYPE Type, CString Name, int Level);
	bool RemoveAttribute (CString Name);
	bool RemoveLevel(int Level);

	int GetAttributeCount();

	void ResetValues();

	void ClearAll();
	
	bool SetAttrib(CString Name, double Value);
	bool SetAttrib(CString Name, bool Value);
	bool SetAttrib(CString Name, CString Value);

	CString GetAttrib(CString Name);
	CString GetAttribNameByNumber(int Which);

	RETURNTYPE GetAttribType(CString Name);

	double  GetAttribF(CString Name);
	bool    GetAttribB(CString Name);
	CString GetAttribS(CString Name);
	
protected:
	struct ATTRIBUTE *AttribTable;
	struct ATTRIBUTE *LastInList;
	struct ATTRIBUTE *FindAttribute(CString Name);
	void DeleteAttribute(struct ATTRIBUTE *Where);
	int TotalAttributes;

};

////////////////////////////////////////////////////////////////
// Class CFunctionTable
// Encapsulates user-functions (See Struct FUNCTIONNODE) and 
// methods to declare, manipulate, and return them and their
// parameters.
class CFunctionTable
{
public:	
	CFunctionTable();
	~CFunctionTable();

	bool AddFunction(CString Name, enum RETURNTYPE Type);
	bool Compile(CString Name, CString SourceCode);
	bool TableValid();
	bool AddParameter(CString FunctName, CString ParmName, enum RETURNTYPE Type);
	bool FindFunction(CString FunctName);
	CString GetFunctNameByNumber(int Which);
	int  GetParameterCount(CString FunctName);
	enum RETURNTYPE GetType(CString FunctName);
	void ClearAll();

	CString GetFunctionName(int WhichFunct);
	CGPProgram * GetFunctionProgram(CString Name);
	CString GetParmName(CString FunctName, int WhichParm);
	enum RETURNTYPE GetParmType(CString FunctName, int WhichParm);

protected:
	struct FUNCTIONNODE * FindFunctionNode(CString Name);
	struct FUNCTIONNODE * Functions;
};

////////////////////////////////////////////////////////////////
// Class CFunctionSelector
// Encapsulates selected functions (user-created and built-ins)
// so as to easily select any random function during genetic
// operations.
class CFunctionSelector
{
public:
	CFunctionSelector();
	~CFunctionSelector();

	CString GetRandomFunction();
	bool IsUserFunction(CString Name);
	bool AddFunction(CString Name);
	void ClearAll();

protected:
	struct FUNCSELECTIONNODE * Selections;
	int TotalSelections;

};

#endif
