/////////////////////////////////////////
// CGPProgram Implemenation File
/////////////////////////////////////////

#include "stdafx.h"
#include "GAPS2.h"
#include "CToken.h"
#include "CSubTree.h"
#include "CGPProgram.h"

#include <cmath>

bool IsDigits(CString & InPiece);
bool IsNumber(CString & TestStr );
bool IsAttribute(CString & TestStr);
bool IsFunctionForm(CString &TestStr);

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
// Constructor and Destructor
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////

CGPProgram::CGPProgram()
{
	ProgramTree = NULL;
	AttribTable = NULL;
	FunctionTable = NULL;

	CurrentFunctLevel = 0;
	MaximumEvaluations = 0;
	CurrentEvaluations = 0;
}

CGPProgram::CGPProgram(CString &NameIn)
{
	m_Name = NameIn;
	CGPProgram();
}

CGPProgram::~CGPProgram()
{
	// Delete function tree (if exists)
	if (ProgramTree != NULL)
		delete ProgramTree;
}


void CGPProgram::SetReturnType(RETURNTYPE InType)
{
	ReturnType = InType;
}

RETURNTYPE CGPProgram::GetReturnType()
{
	// Returns the return value type of the program
	return ReturnType;
}


/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
// Token Stream 
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
// Need to get each token from the source stream.
// Thus, we need to be able to grab the tokens as they come off the input stream.
/////////////////////////////////////////////////////////////////////////////////
void CGPProgram::GetNextToken()
{
	// Attributes
	// Single-char tokens include:
	//         { } ( ) , ; " + - * / %  
	// All-Alphabetic strings are any that do not start with a command
	// and are contiguous.
	// Numbers
	if (ParseString.GetLength() == 0) 
	{
		CurrentToken = CString("");
		return;
	}

	// Check for single-char tokens
	if (ParseString.Left(1) == CString("{") ||
		ParseString.Left(1) == CString("}") ||
		ParseString.Left(1) == CString("(") ||
		ParseString.Left(1) == CString(")") ||
		ParseString.Left(1) == CString(",") ||
		ParseString.Left(1) == CString(";") ||
		ParseString.Left(1) == CString("\"") ||
		ParseString.Left(1) == CString("+") ||
		ParseString.Left(1) == CString("-") ||
		ParseString.Left(1) == CString("*") ||
		ParseString.Left(1) == CString("/") ||
		ParseString.Left(1) == CString("%") ||
		ParseString.Left(1) == CString("\""))
	{
		CurrentToken = ParseString.Left(1);
		ParseString.Delete(0);
		return;
	}

	// Check for two-byte Assignment operator
	if (ParseString.Left(2) == CString(":="))
	{
		CurrentToken = ParseString.Left(2);
		ParseString.Delete(0, 2);
		return;
	}

	// Check for two-byte comparator tokens
	if (ParseString.Left(2) == CString(">=") ||
		ParseString.Left(2) == CString("<=") ||
		ParseString.Left(2) == CString("<>") ||
		ParseString.Left(2) == CString("!="))
	{
		CurrentToken = ParseString.Left(2);
		ParseString.Delete(0, 2);
		return;
	}

	// Check for one-byte comparator tokens
	if (ParseString.Left(1) == CString(">") ||
		ParseString.Left(1) == CString("<") ||
		ParseString.Left(1) == CString("="))
	{
		CurrentToken = ParseString.Left(1);
		ParseString.Delete(0);
		return;
	}

	// Check for 'and', 'or', 'xor', and 'not'
	if (ParseString.Left(3) == CString("not"))
	{
		CurrentToken = ParseString.Left(3);
		ParseString.Delete(0, 3);
		return;
	}

	if (ParseString.Left(2) == CString("||") ||
		ParseString.Left(2) == CString("&&") ||
		ParseString.Left(2) == CString("!|"))
	{
		CurrentToken = ParseString.Left(2);
		ParseString.Delete(0, 2);
		return;
	}

	// Check for Boolean Constants
	if (ParseString.Left(4) == CString("TRUE"))
	{
		CurrentToken = ParseString.Left(4);
		ParseString.Delete(0, 4);
		return;
	}

	if (ParseString.Left(5) == CString("FALSE"))
	{
		CurrentToken = ParseString.Left(5);
		ParseString.Delete(0, 5);
		return;
	}

	if (ParseString.Left(5) == CString("Value"))
	{
		CurrentToken = ParseString.Left(5);
		ParseString.Delete(0, 5);
		return;
	}

	// Attribute
	if (ParseString.Left(1) == CString("&"))
	{
		CurrentToken = ParseString.Left(1);
		ParseString.Delete(0);

		if ((ParseString.GetAt(0) >= 'a' && ParseString.GetAt(0) <= 'z') ||
			(ParseString.GetAt(0) >= 'A' && ParseString.GetAt(0) <= 'Z'))
		{
			CurrentToken += ParseString.Left(1);
			ParseString.Delete(0);
			if (ParseString.GetLength() == 0)
				return;
		}

		while ( 
			(ParseString.GetAt(0) >= 'a' && ParseString.GetAt(0) <= 'z') ||
			(ParseString.GetAt(0) >= 'A' && ParseString.GetAt(0) <= 'Z') ||
			(ParseString.GetAt(0) >= '0' && ParseString.GetAt(0) <= '9'))
		{
			CurrentToken += ParseString.Left(1);
			ParseString.Delete(0);
			if (ParseString.GetLength() == 0)
				return;
		}

		return;
	}

	// Command or Function
	if ((ParseString.GetAt(0) >= 'a' && ParseString.GetAt(0) <= 'z') ||
		(ParseString.GetAt(0) >= 'A' && ParseString.GetAt(0) <= 'Z'))
	{
		CurrentToken = ParseString.Left(1);
		ParseString.Delete(0);
		if (ParseString.GetLength() == 0)
			return;

		while ( 
			(ParseString.GetAt(0) >= 'a' && ParseString.GetAt(0) <= 'z') ||
			(ParseString.GetAt(0) >= 'A' && ParseString.GetAt(0) <= 'Z') ||
			(ParseString.GetAt(0) >= '0' && ParseString.GetAt(0) <= '9'))
		{
			CurrentToken += ParseString.Left(1);
			ParseString.Delete(0);
			if (ParseString.GetLength() == 0)
				return;
		}

		return;
	}

	// Number
	if (ParseString.GetAt(0) >= '0' && ParseString.GetAt(0) <= '9')
	{
		bool DecimalPoint = FALSE;
		bool Mantissa     = FALSE;

		CurrentToken = ParseString.Left(1);
		ParseString.Delete(0);

		// <Number>            ::= D.D
		//                       | D.DeD
		//                       | D.De-D
		// <D>                 ::= d<D> | NULL
		// d                   ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
		
		if (ParseString.GetLength() == 0)
			return;
		while (ParseString.GetLength() > 0 &&
			ParseString.GetAt(0) >= '0' && ParseString.GetAt(0) <= '9')
		{
			CurrentToken += ParseString.Left(1);
			ParseString.Delete(0);
		}

		if (ParseString.GetLength() > 0)
		{
			if (ParseString.GetAt(0) == '.')
			{
				CurrentToken += ParseString.Left(1);
				ParseString.Delete(0);

				while (ParseString.GetAt(0) >= '0' && ParseString.GetAt(0) <= '9')
				{
					CurrentToken += ParseString.Left(1);
					ParseString.Delete(0);
					if (ParseString.GetLength() == 0)
						return;
				}
			}
		}

		if (ParseString.GetLength() > 0)
		{
			if (ParseString.GetAt(0) == 'e')
			{
				CurrentToken += ParseString.Left(1);
				ParseString.Delete(0);

				if (ParseString.GetAt(0) == '-')
				{
					CurrentToken += ParseString.Left(1);
					ParseString.Delete(0);
					if (ParseString.GetLength() == 0)
						return;
				}

				while (ParseString.GetAt(0) >= '0' && ParseString.GetAt(0) <= '9')
				{
					CurrentToken += ParseString.Left(1);
					ParseString.Delete(0);
					if (ParseString.GetLength() == 0)
						return;
				}
			}
			return;
		}
	}
}


void CGPProgram::SetRPBTree(CSubTree * NewRPBTree)
{
	RPBTree = NewRPBTree;
}

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
// Main Compile Commands
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
bool CGPProgram::Compile (CString SourceIn)
{
	m_SourceCode = SourceIn;

	try 
	{
		return Compile();
	}
	catch (CompilationException CompileError)
	{
		AfxMessageBox(CString("Error in compile: ") + CompileError.GetError());
 		return FALSE;
	}
}

bool CGPProgram::Compile ()
{

	ParseString = m_SourceCode;

	if (ProgramTree != NULL)
		delete ProgramTree;


	// Delete Whitespace from non-string constants
	int index = 0;
	bool InStrConst = FALSE;
	
	while (index < ParseString.GetLength())
	{
		if (ParseString.GetAt(index) == '\"')
			InStrConst = !InStrConst;
		else
		{
			if (!InStrConst)
			{
				if (ParseString.GetAt(index) == 10 ||
					ParseString.GetAt(index) == 13 ||
					ParseString.GetAt(index) == 32 ||
					ParseString.GetAt(index) == 9)
				{
					ParseString.Delete(index);
					continue;
				}
			}
		}
		index++;
	}

	ProgramTree = CompileProgram();
	return (ProgramTree != NULL);
}


/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
// <Program>
// Start of Grammar Implementation
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
// <Program>           ::= <ComplexStatement>
//                       | <Statements>
//                       | <Expression>
CSubTree * CGPProgram::CompileProgram()
{
	CSubTree * ReturnTree = NULL;

	GetNextToken();

	if (CurrentToken.GetLength() == 0)
		return NULL;

	CString KeepParseStr = ParseString;
	CString KeepToken    = CurrentToken;
	
	ReturnTree = CompileExpression();
	if (ReturnTree == NULL || CurrentToken == CString(";") || ParseString.GetLength() != 0)
	{
		if (CurrentToken = CString(";") || ParseString.GetLength() != 0)
			delete ReturnTree;
		ReturnTree = NULL;
		ParseString = KeepParseStr;
		CurrentToken = KeepToken;
		if ((ReturnTree = CompileStatements()) == NULL ||
			ParseString.GetLength() != 0)
			return NULL;
	}

	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
// <ComplexStatement> ::=  {<Statements>}
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileComplexStatement()
{
	CSubTree * CurrTree = NULL;

	// {<Statements>}
	if (CurrentToken == CString("{"))
	{
		GetNextToken();

		if ((CurrTree = CompileStatements()) == NULL)
			return NULL;
		if (CurrentToken != CString("}"))
		{
			delete CurrTree;
			throw CompilationException(CString("Expected '}'"));
			// ERROR Condition - Expected "}"
			return NULL;
		}

		GetNextToken();
		return CurrTree;
	}

	return NULL;
}

/////////////////////////////////////////////////////////////////////////////////
// <Statements>        ::= <Statement> 
//                       | <Statement> ; <Statements>  
//                       | <ComplexStatement>
//                       | NULL
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileStatements()
{
	CSubTree * ReturnTree = NULL;
	CSubTree * CurrTree = NULL;

	//| <ComplexStatement>
	if ((ReturnTree = CompileComplexStatement()) != NULL)
		return ReturnTree;

	// <Statement> ; <Statements>
	// <Statement> ; 

	ReturnTree = CompileStatement();
	if (ParseString.GetLength() > 0)
	{
		if (ReturnTree != NULL)
		{
			if (CurrentToken == CString("}"))
			{
				return ReturnTree;
			}
			CurrTree = CompileStatements();
			if (CurrTree != NULL)
			{
				ReturnTree->SetNextChild(CurrTree);
				return ReturnTree;
			}

			delete ReturnTree;
			return NULL;
		}
	}

	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
// <Statement>         ::= <Attribute> := <Expression>
//                       | Map ( <Attribute> ) { <ValueExprs> }
//                       | If ( <BooleanExpr> ) <ComplexStatement>
//                       | If ( <BooleanExpr> ) <ComplexStatement> else <ComplexStatement>
//                       | While ( <BooleanExpr> ) <ComplexStatement>
//                       | DoUntil( <BooleanExpr> ) <ComplexStatement>
//                       | Push( <Attribute> , <Expression> )
//                       | Pop( <Attribute> )
//                       | Replace( <Attribute> , <StringExpression> , <NumExpression> , <NumExpression> )
/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileStatement()
{
	CSubTree * CurrTree;

	// <Attribute> := <Expression>
	if (IsAttribute(CurrentToken))
	{
		CurrTree = CompileAssign();
		if (CurrTree == NULL)
		{
			// Error Condition -- Expecting valid binary assign statement
			return NULL;
		}

		if (CurrentToken != CString(";"))
		{
			delete CurrTree;
			throw CompilationException(CString("Expected ';'"));
			// Error Condition -- Expected ";"
			return NULL;
		}

		GetNextToken();

		return CurrTree;
		
	}

	// Map ( <Attribute> ) { <ValueExprs> }
	if (CurrentToken == CString("Map"))
	{
		CurrTree = CompileMap();
		if (CurrTree == NULL)
		{
			// Error Condition - Map Statement Invalid
			return NULL;
		}
		return CurrTree;
	}

	// If ( <BooleanExpr> ) <ComplexStatement> else <ComplexStatement>
	// If ( <BooleanExpr> ) <ComplexStatement>
	if (CurrentToken == CString("If"))
	{
		CurrTree = CompileIf();
		if (CurrTree == NULL)
		{
			// Error Condition - If Statement invalid
			return NULL;
		}
		return CurrTree;
	}

	// While ( <BooleanExpr> ) <ComplexStatement>
	if (CurrentToken == CString("While"))
	{
		CurrTree = CompileWhile();
		if (CurrTree == NULL)
		{
			// Error Condition - While Statement invalid
			return NULL;
		}
		return CurrTree;
	}

	// DoUntil( <BooleanExpr> ) <ComplexStatement>
	if (CurrentToken == CString("DoUntil"))
	{
		CurrTree = CompileDoUntil();
		if (CurrTree == NULL)
		{
			// Error Condition - DoUntil Statement invalid
			return NULL;
		}
		return CurrTree;
	}

	// Push( <Attribute> , <Expression> )
	if (CurrentToken == CString("Push"))
	{
		CurrTree = CompilePush();
		if (CurrTree == NULL)
		{
			// Error Condition - Push Statement invalid
			return NULL;
		}
		if (CurrentToken != CString(";"))
		{
			delete CurrTree;
			// Error Condition -- Expected ";"
			return NULL;
		}

		GetNextToken();

		return CurrTree;
	}

	// Pop( <Attribute> )
	if (CurrentToken == CString("Pop"))
	{
		CurrTree = CompilePop();
		if (CurrTree == NULL)
		{
			// Error Condition - Pop Statement invalid
			return NULL;
		}
		if (CurrentToken != CString(";"))
		{
			delete CurrTree;
			// Error Condition -- Expected ";"
			return NULL;
		}

		GetNextToken();

		return CurrTree;
	}

	// Replace( <Attribute> , <StringExpression> , <NumExpression> , <NumExpression> )
	if (CurrentToken == CString("Replace"))
	{
		CurrTree = CompileReplace();
		if (CurrTree == NULL)
		{
			// Error Condition - Pop Statement invalid
			return NULL;
		}
		if (CurrentToken != CString(";"))
		{
			delete CurrTree;
			// Error Condition -- Expected ";"
			return NULL;
		}

		GetNextToken();
		return CurrTree;
	}

	if (CurrentToken == CString("Declare"))
	{
		CurrTree = CompileDeclare();
		if (CurrTree == NULL)
		{
			// Error - Bad Declare
			return NULL;
		}
		if (CurrentToken != CString(";"))
		{
			delete CurrTree;
			// Error Condition -- Expected ";"
			return NULL;
		}

		GetNextToken();
		return CurrTree;
		
	}

	// Try a User Function
	CurrTree = CompileUserFunction();
	if (CurrTree != NULL)
	{
		if (CurrentToken != CString(";"))
		{
			delete CurrTree;
			return NULL;
		}

		GetNextToken();
		return CurrTree;
	}

	return NULL;
}

/////////////////////////////////////////////////////////////////////////////////
// <Attribute> := <Expression> 
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileAssign()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	CurrTree = CompileAttribute();
	if (CurrTree == NULL)
	{
		// Error - Bad Attribute form
		return NULL;
	}
	
	if (CurrentToken != CString(":="))
	{
		delete CurrTree;  // Not necessarily an error
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNASSIGN, CString("Assign"));
	ReturnTree->m_NumArguments = 2;
	ReturnTree->SetChild(0, CurrTree);

	CurrTree = CompileExpression();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Invalid Expression in Assign statement"));
		// Throw an exception   -- ERROR condition: Expression is invalid
		return NULL;
	}

	ReturnTree->SetChild(1, CurrTree);
	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
// Map ( <Attribute> ) { <ValueExprs> }
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileMap()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();

	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' in Map statement"));
		// ERROR condition - Expected "("
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNMAP, CString("Map"));
	ReturnTree->m_NumArguments = 2;
	
	CurrTree = CompileAttribute();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected valid Attribute (&<attr> format)"));
		return NULL;
	}
	ReturnTree->SetChild(0, CurrTree);

	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' in Map Statement"));
		return NULL;
	}

	GetNextToken();
	if (CurrentToken != CString("{"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected '{' in Map Statement"));
		return NULL;
	}
	
	GetNextToken();
	
	CurrTree = CompileValueExprs();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		//               ERROR Condition: Value Expressions invalid
		return NULL;
	}

	if (CurrentToken != CString("}"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected '}'in Map Statement"));
		return NULL;
	}
	ReturnTree->SetChild(1, CurrTree);

	GetNextToken();

	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
// <ValueExprs> ::= <Values> <ValueExprs> | NULL
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileValueExprs()
{
	CSubTree * ReturnTree;

	ReturnTree = CompileValues();
	if (CurrentToken != CString("}") && ReturnTree != NULL) 
	{
		ReturnTree->SetChild(2, CompileValueExprs());
	}
	
	return ReturnTree;
}

// <Values> ::= Value <Expression> <ComplexStatement>
CSubTree * CGPProgram::CompileValues()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	if (CurrentToken != CString("Value"))
	{
		throw CompilationException(CString("Expected 'Value'"));
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNVALUE, CString("Value"));
	ReturnTree->m_NumArguments = 2;

	CurrTree = CompileExpression();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected valid expression in Value statement"));
		return NULL;
	}
	ReturnTree->SetChild(0, CurrTree);

	CurrTree = CompileComplexStatement();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		// ERROR: Complex Statement
		return NULL;
	}

	ReturnTree->SetChild(1, CurrTree);
	
	return ReturnTree;

}


/////////////////////////////////////////////////////////////////////////////////
// If ( <BooleanExpr> ) <ComplexStatement> else <ComplexStatement>
// If ( <BooleanExpr> ) <ComplexStatement>
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileIf()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();
	
	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' in If statement"));
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNIF, CString("If"));
	ReturnTree->m_NumArguments = 3;

//	CurrTree = CompileTopLevelExpression();
	CurrTree = CompileExpression();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		// ERROR - if statement not a boolean expression
		return NULL;
	}

	ReturnTree->SetChild(0, CurrTree);
	
	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' in If Statement"));
		return NULL;
	}

	GetNextToken();

	CurrTree = CompileComplexStatement();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		// Error in If Statement
		return NULL;
	}

	ReturnTree->SetChild(1, CurrTree);

	if (CurrentToken == CString("Else"))
	{
		GetNextToken();
		CurrTree = CompileComplexStatement();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			// Error - Else clause cannot refer to a null statement
			return NULL;
		}

		ReturnTree->SetChild(2, CurrTree);

		return ReturnTree;
	}

	// No Else clause, but something needs to be there.
	CurrTree = new CSubTree(ttFNNOOP, CString("No-op"));
	ReturnTree->SetChild(2, CurrTree);

	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
// While ( <BooleanExpr> ) <ComplexStatement>
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileWhile()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();
	
	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' after While statement"));
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNWHILE, CString("While"));
	CurrTree = CompileTopLevelExpression();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		// ERROR - if statement not a boolean expression
		return NULL;
	}

	ReturnTree->SetChild(0, CurrTree);
	
	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' after While statement"));
		return NULL;
	}

	GetNextToken();

	CurrTree = CompileComplexStatement();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected complex statement after While statement"));
		return NULL;
	}

	ReturnTree->SetChild(1, CurrTree);
	ReturnTree->m_NumArguments = 2;

	return ReturnTree;
}
	
/////////////////////////////////////////////////////////////////////////////////
// DoUntil( <BooleanExpr> ) <ComplexStatement>
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileDoUntil()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();
	
	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' after DoUntil"));
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNWHILE, CString("Do-Until"));
	CurrTree = CompileTopLevelExpression();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		// ERROR - if statement not a boolean expression
		return NULL;
	}

	ReturnTree->SetChild(1, CurrTree);
	
	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' after expression in DoUntil"));
		return NULL;
	}

	GetNextToken();

	CurrTree = CompileComplexStatement();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected complex statement after DoUntil"));
		return NULL;
	}

	ReturnTree->SetChild(0, CurrTree);
	ReturnTree->m_NumArguments = 2;

	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
// Push( <Attribute> , <Expression> )
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompilePush()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();

	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' after Push statement"));
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNPUSH, CString("Push"));
	ReturnTree->m_NumArguments = 2;

	CurrTree = CompileAttribute();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("First argument of Push is not an Attribute "));
		return NULL;
	}

	if (CurrentToken != CString(","))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ',' in Push statement"));
		return NULL;
	}
	ReturnTree->SetChild(0, CurrTree);
	GetNextToken();

	CurrTree = CompileExpression();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected valid expression as second argument of Push"));
		return NULL;
	}

	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' after Push statement"));
		return NULL;
	}

	ReturnTree->SetChild(1, CurrTree);
	GetNextToken();
	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
// Pop( <Attribute> )
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompilePop()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();

	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' after Pop statement"));
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNPOP, CString("Pop"));

	CurrTree = CompileAttribute();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected attribute as first argument of Pop statement"));
		return NULL;
	}

	ReturnTree->SetChild(0, CurrTree);

	if (CurrentToken != CString(","))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ',' after first Pop argument"));
		return NULL;
	}

	GetNextToken();

	CurrTree = CompileAttribute();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected attribute as second argument of Pop statement"));
		return NULL;
	}

	ReturnTree->SetChild(1, CurrTree);

	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' after second argument of Pop statement"));
		return NULL;
	}

	GetNextToken();
	ReturnTree->m_NumArguments = 2;
	return ReturnTree;
}

/////////////////////////////////////////////////////////////////////////////////
// Replace( <Attribute> , <StringExpression> , <NumExpression> , <NumExpression> )
/////////////////////////////////////////////////////////////////////////////////
CSubTree * CGPProgram::CompileReplace()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();

	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' after Replace statement"));
		return NULL;
	}

	GetNextToken();

	ReturnTree = new CSubTree(ttFNREPLACE, CString("Replace"));

	CurrTree = CompileAttribute();
	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected valid attribute as first argument of Replace statement"));
		return NULL;
	}

	if (CurrentToken != CString(","))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ',' after first argument of Replace statement"));
		return NULL;
	}
	ReturnTree->SetChild(0, CurrTree);
	GetNextToken();
	
	CurrTree = CompileStringExpr();

	if (CurrTree == NULL)
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected valid string argument in Replace statement"));
	}

	if (CurrentToken != CString(","))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ',' after second argument in Replace statement"));
		return NULL;
	}
	ReturnTree->SetChild(1, CurrTree);
	GetNextToken();

	CurrTree = CompileNumExpression();
	if (CurrentToken != CString(","))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ',' after third argument in Replace statement"));
		return NULL;
	}
	ReturnTree->SetChild(2, CurrTree);
	GetNextToken();

	CurrTree = CompileNumExpression();
	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' after Replace statement"));
		return NULL;
	}

	ReturnTree->SetChild(3, CurrTree);
	GetNextToken();
	ReturnTree->m_NumArguments = 4;
	return ReturnTree;
}

// Declare (<Attribute>, [Number | Boolean | String])
CSubTree * CGPProgram::CompileDeclare()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	GetNextToken();
	if (CurrentToken != CString("("))
	{
		throw CompilationException(CString("Expected '(' after Declare statement"));
		return NULL;
	}

	GetNextToken();
	CurrTree = CompileAttribute();

	if (CurrTree == NULL)
	{
		throw CompilationException(CString("Expected valid Attribute in Declare statement"));
		return NULL;
	}

	if (CurrentToken != CString(","))
	{
		delete CurrTree;
		throw CompilationException(CString("Expected ',' in Declare statement"));
		return NULL;
	}

	ReturnTree = new CSubTree(ttFNDECLARE, CString("Declare"));
	ReturnTree->m_NumArguments = 2;

	ReturnTree->SetChild(0, CurrTree);

	GetNextToken();

	if (CurrentToken != CString("\""))
	{
		delete ReturnTree;
		throw CompilationException(CString("Declare's types need to be enclosed in quotation marks"));
		return NULL;  // 
	}

	GetNextToken();

	if (CurrentToken != CString("Number") &&
		CurrentToken != CString("Boolean") &&
		CurrentToken != CString("String"))
	{
		delete ReturnTree;
		throw CompilationException(CString("bad type declaration in Declare: ") + CurrentToken);
		return NULL;  // 
	}
	
	CurrTree = new CSubTree(ttSTRING, CurrentToken);

	ReturnTree->SetChild(1, CurrTree);

	GetNextToken();


	if (CurrentToken != CString("\""))
	{
		delete ReturnTree;
		throw CompilationException(CString("Declare's types need to be enclosed in quotation marks"));
		return NULL;  // 
	}

	GetNextToken();

	if (CurrentToken != CString(")"))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected ')' after Declare statement"));
		return NULL;
	}
	GetNextToken();
	return ReturnTree;

}

CSubTree * CGPProgram::CompileAttribute()
{
	CSubTree * ReturnTree;

	if (CurrentToken.Left(1) != CString("&"))
	{
		return NULL;
	}


	// Grab the attribute name, peeling off the leftmost '&'
	ReturnTree = new CSubTree(ttUSRATTRIB, CurrentToken.Mid(1));
	ReturnTree->m_NumArguments = 0;


	GetNextToken();

	return ReturnTree;
}

// <Expression>        ::= <TopLevelExpression>
//                       | " <StringExpression> "
CSubTree * CGPProgram::CompileExpression()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;
	CString KeepParseStr = ParseString;
	CString KeepToken    = CurrentToken;

	CurrTree = CompileStringExpr();
	// May be part of an equality
	if (CurrentToken == CString("="))
	{
		GetNextToken();
		ReturnTree = new CSubTree(ttCOMPAREOP, CString("="));
		ReturnTree->SetChild(0, CurrTree);

		CurrTree = CompileStringExpr();
		if (CurrTree != NULL)
		{
			ReturnTree->SetChild(1, CurrTree);
			CurrTree = ReturnTree;
		}
	}
	if (CurrTree != NULL)
		return CurrTree;

	ParseString = KeepParseStr;
	CurrentToken = KeepToken;
	CurrTree = CompileTopLevelExpression();
	if (CurrTree != NULL)
		return CurrTree;

	return NULL;
}

// <TopLevelExpression>::= not <TopLevelExpression>
//                       | <FullTerm> [|| <FullTerm>]...
CSubTree * CGPProgram::CompileTopLevelExpression()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;
	CString KeepParseStr = ParseString;
	CString KeepToken    = CurrentToken;

	// not <TopLevelExpression>
	if (CurrentToken == CString("not"))
	{
		ReturnTree = new CSubTree(ttNOT, CString("not"));
		ReturnTree->m_NumArguments = 1;

		GetNextToken();

		CurrTree = CompileTopLevelExpression();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			return NULL;
		}

		ReturnTree->SetChild(0, CurrTree);
		return ReturnTree;
	}

	// <FullTerm> [|| <TopLevelExpression>]...
	CurrTree = CompileFullTerm();
	
	while (CurrentToken == CString("||") && CurrTree != NULL)
	{
		ReturnTree = new CSubTree(ttOR, CString("||"));
		ReturnTree->m_NumArguments = 2;
		
		GetNextToken();

		ReturnTree->SetChild(0, CurrTree);

		CurrTree = CompileTopLevelExpression();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			return NULL;
		}
		
		ReturnTree->SetChild(1, CurrTree);
		CurrTree = ReturnTree;
	}

	return CurrTree;
}


// <FullTerm>          ::= <FullFactor> [&& <TopLevelExpression>]...
CSubTree * CGPProgram::CompileFullTerm()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	CurrTree = CompileFullFactor();
	while (CurrentToken == CString("&&") && CurrTree != NULL)
	{
		ReturnTree = new CSubTree(ttAND, CString("&&"));
		ReturnTree->m_NumArguments = 2;

		GetNextToken();

		ReturnTree->SetChild(0, CurrTree);

		CurrTree = CompileTopLevelExpression();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			return NULL;
		}

		ReturnTree->SetChild(1, CurrTree);
		CurrTree = ReturnTree;
	}


	return CurrTree;
}

// <FullFactor>        ::= <NumExpression> [[< | > | != | <> | >= | <= | =] <TopLevelExpression>]...
CSubTree * CGPProgram::CompileFullFactor()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;


	CurrTree = CompileNumExpression();

	while (CurrTree != NULL &&
		(CurrentToken == CString("<") || 
		 CurrentToken == CString(">") || 
		 CurrentToken == CString("!=") || 
		 CurrentToken == CString("<>") || 
		 CurrentToken == CString(">=") || 
		 CurrentToken == CString("<=") || 
		 CurrentToken == CString("=")))
	{
		ReturnTree = new CSubTree(ttCOMPAREOP, CurrentToken);
		ReturnTree->m_NumArguments = 2;

		GetNextToken();
		ReturnTree->SetChild(0, CurrTree);

		CurrTree = CompileTopLevelExpression();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			// Error - Expecting Numeric expression 
			return NULL;
		}

		ReturnTree->SetChild(1, CurrTree);
		CurrTree = ReturnTree;

	}

	
	return CurrTree;
}


// <NumExpression>     ::= <Term> [[+ | -] <TopLevelExpression>]...

CSubTree * CGPProgram::CompileNumExpression()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;
	CString KeepParseStr = ParseString;
	CString KeepToken    = CurrentToken;

	if (CurrentToken.GetLength() == 0)
		return NULL;


	// <Term> [[+ | -] <Term>]...
	// Note: Breaking down +/- by terms establishes operator precedence
	ParseString  = KeepParseStr;
	CurrentToken = KeepToken;
	CurrTree = CompileTerm();

	while (CurrentToken.GetLength() > 0 &&
		   (CurrentToken.Left(1) == "-" ||
		    CurrentToken.Left(1) == "+"))
	{
		if (CurrTree != NULL)
		{

			ReturnTree = new CSubTree(CurrentToken.GetAt(0) == '+' ? ttADD : ttSUBTRACT, CurrentToken);
			ReturnTree->SetChild(0, CurrTree);
			ReturnTree->m_NumArguments = 2;

			GetNextToken();
 			CurrTree = CompileTopLevelExpression();
			if (CurrTree == NULL)
			{
				delete ReturnTree;
				throw CompilationException(CString("Invalid numeric expression after +/- token"));
				return NULL;
			}
			
			ReturnTree->SetChild(1, CurrTree);
			CurrTree = ReturnTree;
		}
	}
	return CurrTree;
}


// <Term>              ::= <Factor> [[* | / |  %] <Factor>] ...
CSubTree * CGPProgram::CompileTerm()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;
	CString KeepParseStr = ParseString;
	CString KeepToken    = CurrentToken;

	if (CurrentToken.GetLength() == 0)
		return NULL;

	CurrTree = CompileFactor();
	
	while (CurrentToken.GetLength() > 0 &&
		(CurrentToken.Left(1) == "*" ||
		CurrentToken.Left(1) == "/" ||
		CurrentToken.GetAt(0) == '%')
		&& CurrTree != NULL)
	{
		if (CurrTree != NULL)
		{
			ReturnTree = new CSubTree;
			if (CurrentToken == CString("*"))
				ReturnTree->SetType(ttMULTIPLY);
			if (CurrentToken == CString("/"))
				ReturnTree->SetType(ttDIVIDE);
			if (CurrentToken.GetAt(0) == '%')
				ReturnTree->SetType(ttMOD);
			ReturnTree->SetVal(CurrentToken);
			ReturnTree->m_NumArguments = 2;

			ReturnTree->SetChild(0, CurrTree);

			GetNextToken();
			CurrTree = CompileFactor();
			if (CurrTree == NULL)
			{
				delete ReturnTree;
				// Error - Right value expects numeric expression
				return NULL;
			}
			
			ReturnTree->SetChild(1, CurrTree);
			CurrTree = ReturnTree;
		}
	}

	if (CurrTree != NULL)
		return CurrTree;
	// Reset parse string back to original state
	ParseString = KeepParseStr;
	CurrentToken = KeepToken;

	return FALSE;
}

// <Factor>            ::= ( <TopLevelExpression> )
//                       | <Attribute>
//                       | - <Number>
//                       | <Number>
//                       | <MathUnaryFunction> ( <NumExpression> )
//                       | TRUE | FALSE
//                       | <UserFunction> ( [<Attribute> [, <Attribute>]...])
CSubTree * CGPProgram::CompileFactor()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;
	CString KeepParseStr = ParseString;
	CString KeepToken    = CurrentToken;

	if (CurrentToken.GetLength() == 0)
		return NULL;

	// ( <TopLevelExpression> )
	if (CurrentToken == CString("("))
	{
		GetNextToken();

		CurrTree = CompileTopLevelExpression();
		if (CurrTree == NULL)
		{
			// Error - Expecting Expression
			return NULL;
		}

		if (CurrentToken != CString(")"))
		{
			delete CurrTree;
			// Error - Mismatched parentheses
			return NULL;
		}
		
		GetNextToken();
		return CurrTree;
	}

	// <Attribute>
	if (IsAttribute(CurrentToken))
	{
		ReturnTree = new CSubTree(ttUSRATTRIB, CurrentToken.Mid(1));
		ReturnTree->m_NumArguments = 0;

		GetNextToken();
		return ReturnTree;
	}

	// - <Number>
	CurrTree = NULL;
	if (CurrentToken == CString("-"))
	{
		CurrTree = new CSubTree(ttNEGATIVE, CString("-"));
		CurrTree->m_NumArguments = 1;
		GetNextToken();
	}

	// <Number>
	if (IsNumber(CurrentToken))
	{
		ReturnTree = new CSubTree(ttFLOAT, CurrentToken);
		ReturnTree->m_NumArguments = 0;

		GetNextToken();
		if (CurrTree != NULL)
		{
			CurrTree->SetChild(0, ReturnTree);
			return CurrTree;
		}
		return ReturnTree;
	}

	// <MathUnaryFunction> ( <NumExpression> )
	ReturnTree = CompileNumUnaryFunction();
	if (ReturnTree != NULL)
	{
		if (CurrentToken != CString("("))
		{
			delete ReturnTree;
			throw CompilationException(CString("Expected '(' after unary Math function"));
			return NULL;
		}
		GetNextToken();
		CurrTree = CompileNumExpression();
		if (CurrentToken != CString(")") || CurrTree == NULL)
		{
			delete ReturnTree;
			// Error - Mismatched paretheses or empty parentheses
			return NULL;
		}
		GetNextToken();
		ReturnTree->SetChild(0, CurrTree);
		return ReturnTree;

	}

	// TRUE | FALSE
	if (CurrentToken == CString("TRUE") ||
		CurrentToken == CString("FALSE"))
	{
		ReturnTree = new CSubTree;
		ReturnTree->SetType(ttBOOL);
		if (CurrentToken == CString("TRUE"))
			ReturnTree->SetVal(CString("1"));
		else
			ReturnTree->SetVal(CString("0"));
		ReturnTree->m_NumArguments = 0;

		GetNextToken();
		return ReturnTree;
	}

	// <UserFunction> ( [<Attribute> [, <Attribute>]...])
	if (IsFunctionForm(CurrentToken))
	{
		CurrTree = CompileUserFunction();
		if (CurrTree != NULL)
			return CurrTree;

		ParseString = KeepParseStr;
		CurrentToken = KeepToken;
	}

	return NULL;
}

// <MathUnaryFunction> ::= sin | cos | tan | log | sqrt | round | abs | trunc
CSubTree * CGPProgram::CompileNumUnaryFunction()
{
	CSubTree * ReturnTree;

	if (CurrentToken.GetLength() == 0)
		return NULL;


	if (CurrentToken == CString("sin") ||
		CurrentToken == CString("cos") ||
		CurrentToken == CString("tan") ||
		CurrentToken == CString("log") ||
		CurrentToken == CString("sqrt") ||
		CurrentToken == CString("round") ||
		CurrentToken == CString("trunc") ||
		CurrentToken == CString("abs"))
	{
		ReturnTree = new CSubTree(ttNUMFUNCT, CurrentToken);
		ReturnTree->m_NumArguments = 1;

		GetNextToken();
		return ReturnTree;
	}
	
	return NULL;
}


// <UserFunction>      ::= <CharStartedAlphaNumString>
CSubTree * CGPProgram::CompileUserFunction()
{
	CSubTree * ReturnTree;
	CSubTree * CurrTree;
	CString FunctName;
	int TotalArgs = 0;

	ReturnTree = new CSubTree;
	if (CurrentToken == CString("RPB"))
		ReturnTree->SetType(ttGENETICFUNC);
	else
		ReturnTree->SetType(ttUSRFUNCTION);

	ReturnTree->SetVal(CurrentToken);
	FunctName = CurrentToken;

	GetNextToken();

	if (CurrentToken != CString("("))
	{
		delete ReturnTree;
		throw CompilationException(CString("Expected '(' after function ") + FunctName);
		return NULL;
	}
	
	GetNextToken();

	if (CurrentToken == CString(")"))
	{
		// No arguments is perfectly valid
		GetNextToken();
		return ReturnTree;
	}

	CurrTree = CompileExpression();
	TotalArgs=1;

	while (CurrentToken == CString(","))
	{
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			throw CompilationException(CString("Invalid expression as argument of function ") + FunctName);
			return NULL;
		}
		GetNextToken();
		
		ReturnTree->SetNextChild(CurrTree);
		CurrTree = CompileExpression();
		TotalArgs++;

	}

	if (CurrentToken != CString(")") || CurrTree == NULL)
	{
		delete ReturnTree;
		// Expecting ")"
		return NULL;
	}

	ReturnTree->SetNextChild(CurrTree);
	ReturnTree->m_NumArguments = TotalArgs;

	GetNextToken();
	return ReturnTree;
}


// <StringExpression>  ::= "<A>"
//                       | <Attribute>
//                       | GetAt ( <Attribute> , <NumExpression> , <NumExpression> )
// <A>                 ::= **(Any string of alphanumeric characters)**
// <C>                 ::= c<C> | NULL
// <c>                 ::= {Any one character from a-z or A-Z}
CSubTree * CGPProgram::CompileStringExpr()
{

	CSubTree * ReturnTree;
	CSubTree * CurrTree;

	CString StoreVal = CString("");

	if (CurrentToken == CString("\""))
	{
		// Must pull off the String itself; since nothing inside is a Token, 
		// we cannot call GetNextToken();

		// Pull off the rest of the string until the last "
		while (ParseString.GetLength() > 0 && ParseString.GetAt(0) != '\"')
		{
			StoreVal += ParseString.GetAt(0);
			ParseString.Delete(0);
		}
		if (ParseString.GetLength() == 0) 
		{
			throw CompilationException(CString("Error - mismatched quotes"));
			return NULL;
		}

		// Pull off remaining "
		ParseString.Delete(0);
		GetNextToken();

		// Create Tree Node ************************
		ReturnTree = new CSubTree(ttSTRING, StoreVal);
		ReturnTree->m_NumArguments = 0;

		return ReturnTree;
		
	}

	if (CurrentToken == CString("GetAt"))
	{
		GetNextToken();
		if (CurrentToken != CString("("))
		{
			throw CompilationException(CString("Expected '(' after GetAt function"));
			return NULL;
		}
		ReturnTree = new CSubTree(ttFNGETAT, CString("GetAt"));
		ReturnTree->m_NumArguments = 3;

		GetNextToken();
		CurrTree = CompileAttribute();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			throw CompilationException(CString("Expected valid Attribute as first argument of GetAt"));
			return NULL;
		}

		ReturnTree->SetChild(0, CurrTree);

		if (CurrentToken != CString(","))
		{
			delete ReturnTree;
			throw CompilationException(CString("Expected ',' in GetAt call"));
			return NULL;
		}

		GetNextToken();
		CurrTree = CompileNumExpression();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			// error - expecting numeric expression
			return NULL;
		}

		ReturnTree->SetChild(1, CurrTree);

		if (CurrentToken != CString(","))
		{
			delete ReturnTree;
			throw CompilationException(CString("Expected ',' in GetAt call"));
			return NULL;
		}

		GetNextToken();
		CurrTree = CompileNumExpression();
		if (CurrTree == NULL)
		{
			delete ReturnTree;
			// error - expecting numeric expression
			return NULL;
		}

		ReturnTree->SetChild(2, CurrTree);

		if (CurrentToken != CString(")"))
		{
			delete ReturnTree;
			throw CompilationException(CString("Expected ')' in GetAt call"));
			return NULL;
		}

		GetNextToken();
		return ReturnTree;
	}
	return NULL;
}

class CSubTree * CGPProgram::FindRPB(class CSubTree * StartNode)
{
	int ThisChild;

	if (StartNode->GetType() == ttGENETICFUNC)
		return StartNode;
	
	class CSubTree * ReturnNode;

	for (ThisChild = 0; StartNode->GetChild(ThisChild) != NULL; ThisChild ++)
	{
		ReturnNode = FindRPB(StartNode->GetChild(ThisChild));
		// Determine the type of call that is being generated
		if (ReturnNode != NULL)
			return ReturnNode;
	}
	return NULL;
}

enum RETURNTYPE CGPProgram::GetRPBContext(class CSubTree * StartNode, CFunctionTable * FunctionTblin)
{
	int ThisChild;

	if (StartNode->GetType() == ttGENETICFUNC)
		return rtRPBFOUND;

	enum RETURNTYPE NextNode;

	for (ThisChild = 0; StartNode->GetChild(ThisChild) != NULL; ThisChild ++)
	{
		NextNode = GetRPBContext(StartNode->GetChild(ThisChild), FunctionTblin);
		if (NextNode == rtRPBFOUND)
			return (GetChildContext(StartNode->GetType(), StartNode->GetVal(), ThisChild, FunctionTblin));
		if (NextNode != rtERROR)
			return NextNode;
	}
	return rtERROR;
	
}

enum RETURNTYPE CGPProgram::GetChildContext(TOKENTYPE CallerType, CString CallerName, int WhichChild, CFunctionTable * FunctionTblin)
{
	// Must determine from the caller type and child what the return type is
	switch (CallerType)
	{
	case ttADD:
	case ttSUBTRACT:
	case ttMULTIPLY:
	case ttDIVIDE:
	case ttMOD:
	case ttCOMPAREOP:
	case ttNUMFUNCT:
	case ttFLOAT:
		return rtFLOAT;

	case ttNOT:
	case ttAND:
	case ttOR:
	case ttBOOL:
		return rtBOOL;

	case ttSTRING:
		return rtSTRING;

	case ttFNIF:
		if (WhichChild == 0)
			return rtBOOL;
		return rtVOID;

	case ttFNNOOP:
		return rtVOID;
	
	case ttFNWHILE:
		if (WhichChild == 0)
			return rtBOOL;
		return rtVOID;

	case ttFNDOUNTIL:
		if (WhichChild == 0)
			return rtBOOL;
		return rtVOID;

	case ttFNASSIGN:
		if (WhichChild == 0)
			return rtATTRIBUTE;
		return rtEXPRESSION;

	case ttFNPUSH:
		if (WhichChild == 0)
			return rtATTRIBUTE;
		return rtEXPRESSION;

	case ttFNPOP:
		return rtATTRIBUTE;

	case ttFNREPLACE:
		switch (WhichChild)
		{
		case 0:
			return rtATTRIBUTE;
		case 1:
			return rtSTRING;
		case 2:
		case 3:
			return rtFLOAT;
		}
		break;

	case ttFNGETAT:
		switch (WhichChild)
		{
		case 0:
			return rtATTRIBUTE;
		case 1:
		case 2:
			return rtFLOAT;
		}

	case ttUSRFUNCTION:
		return(FunctionTblin->GetType(CallerName));

	case ttUSRATTRIB:
		return rtATTRIBUTE;
	}

	return rtERROR;
	
}

enum RETURNTYPE CGPProgram::GetBuiltinContext(CString WhichBuiltin)
{
	if (WhichBuiltin == CString("+") ||
		WhichBuiltin == CString("-") ||
		WhichBuiltin == CString("*") ||
		WhichBuiltin == CString("/") ||
		WhichBuiltin == CString("%") ||
		WhichBuiltin == CString("sin") ||
		WhichBuiltin == CString("cos") ||
		WhichBuiltin == CString("tan") ||
		WhichBuiltin == CString("abs") ||
		WhichBuiltin == CString("sqrt") ||
		WhichBuiltin == CString("round") ||
		WhichBuiltin == CString("trunc") ||
		WhichBuiltin == CString("log"))
		return rtFLOAT;

	if (WhichBuiltin == CString("not") ||
		WhichBuiltin == CString("&&") ||
		WhichBuiltin == CString("||") ||
		WhichBuiltin == CString(">") ||
		WhichBuiltin == CString("<") ||
		WhichBuiltin == CString(">=") ||
		WhichBuiltin == CString("=") ||
		WhichBuiltin == CString(">=") ||
		WhichBuiltin == CString("<=") ||
		WhichBuiltin == CString("!="))
		return rtBOOL;

	if (WhichBuiltin == CString("If") ||
		WhichBuiltin == CString("While") ||
		WhichBuiltin == CString("DoUntil") ||
		WhichBuiltin == CString(":=") ||
		WhichBuiltin == CString("Push") ||
		WhichBuiltin == CString("Pop") ||
		WhichBuiltin == CString("Replace") ||
		WhichBuiltin == CString("GetAt"))
		return rtVOID;

	return rtERROR;
}


bool CGPProgram::CheckLegalFunctions(class CSubTree * StartNode, class CFunctionTable * FunctionTblIn)
{
	int ThisChild;

	if (StartNode->GetType() == ttUSRFUNCTION)
	{
		if (!FunctionTblIn->FindFunction(StartNode->GetVal()))
			return FALSE;
	}

	for (ThisChild = 0; StartNode->GetChild(ThisChild) != NULL; ThisChild++)
	{
		if (!CheckLegalFunctions(StartNode->GetChild(ThisChild), FunctionTblIn))
			return FALSE;
	}

	return TRUE;
}

bool CGPProgram::CheckUsedAttributes(class CSubTree * StartNode, struct UserAttributes AttributeTbl[], int TotalAttribs)
{
	int index;
	int ThisAttr;

	if (StartNode->GetType() == ttUSRATTRIB)
	{
		for (ThisAttr = 0; ThisAttr < TotalAttribs; ThisAttr++)
		{
			if (StartNode->GetVal() == AttributeTbl[ThisAttr].Name)
				return TRUE;
		}
		if (ThisAttr == TotalAttribs)
			return FALSE;
	}

	for (index =0 ; index < StartNode->GetChildCount(); index++)
	{
		if (!CheckUsedAttributes(StartNode->GetChild(index), AttributeTbl, TotalAttribs))
			return FALSE;
	}

	return TRUE;
}

CompilationException::CompilationException(CString & ErrorMsg)
{
	OurErrorMessage = ErrorMsg;
}


CompilationException::~CompilationException()
{
}

CString CompilationException::GetError()
{
	return OurErrorMessage;
}

