/////////////////////////////////////////
// CGPProgram Implemenation File
/////////////////////////////////////////

#include "stdafx.h"
#include "GAPS2.h"
#include "CToken.h"
#include "CSubTree.h"
#include "CExecEnv.h"
#include "CGPProgram.h"

#include <cmath>

int IsBoolean(CString &TestStr);
bool IsNumber(CString &TestStr);

///////////////////////////////////////////////////////////////////////
////////////////////////   EXECUTE   //////////////////////////////////
///////////////////////////////////////////////////////////////////////
///////////////     Execution of the Program    ///////////////////////
///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//                                                                   //
// Execute is a recursive program that executes itself and all its   //
// children                                                          //
//                                                                   //
///////////////////////////////////////////////////////////////////////

CString CGPProgram::Execute(class CAttributeTable * AttribTblIn, class CFunctionTable * FunctionTblIn, long MaxEvaluationsAllowed)
{
	CString FinalValue;

	CurrentEvaluations = 0;
	MaximumEvaluations = MaxEvaluationsAllowed;

	AttribTable = AttribTblIn;
	FunctionTable = FunctionTblIn;

	FinalValue = ExecRecursive(ProgramTree);
	AttribTable = NULL;

	return FinalValue;
}

CString CGPProgram::ExecRecursive(class CSubTree * Node)
{
	CString ExprValue;
	CString AttrVal;
	CString RetVal;
	CString ReadValue;
	CString LeftValue;
	CString RightValue;

	CurrentEvaluations++;
	if (MaximumEvaluations != 0)
	{
		if (CurrentEvaluations > MaximumEvaluations)
		{
			return CString("TIMEOUTERROR");
		}
	}

	switch(Node->GetToken()->GetType())
	{
	case ttFNDECLARE:
		// Child 0 is the Attribute
		// Child 1 is String constant of Number, Boolean, or String
		// Child 2 is the Next Statement
		{
			enum RETURNTYPE SetType;
			if (Node->GetChild(1)->GetVal() == CString("Number"))
				SetType = rtFLOAT;
			if (Node->GetChild(1)->GetVal() == CString("Boolean"))
				SetType = rtBOOL;
			if (Node->GetChild(1)->GetVal() == CString("String"))
				SetType = rtSTRING;

			if (!AttribTable->AddAttribute(SetType, Node->GetChild(0)->GetVal(), CurrentFunctLevel))
			{
				// Error condition -- Attribute already declared
			}
		}
		// Execute next statement, if available
		if (Node->GetChild(2) != NULL)
			ExecRecursive(Node->GetChild(2));
		break;

	case ttFNASSIGN:  // Assignment Operator
		// Child 0 is the assigned Attribute 
		// Child 1 is the expression to be assigned
		// Child 2 (if exists) is the next statement
		ReadValue = ExecRecursive(Node->GetChild(1));
		AttribTable->SetAttrib(Node->GetChild(0)->GetVal(), ReadValue);
		// Execute next statement, if available
		if (Node->GetChild(2) != NULL)
			ExecRecursive(Node->GetChild(2));
		break;


	case ttFNIF:
		// Child 0 is the Boolean Expression
		// Child 1 is the 'then' statement(s)
		// Child 2 is the 'else' statement(s) (if exists)
		// Child 3 (if exists) is the next statement
		if (GetBoolExprVal(ExecRecursive(Node->GetChild(0))))
		{
			ExecRecursive(Node->GetChild(1));
			if (Node->GetChild(1) == NULL)
				int DEBUGthis = 1;
		}
		else
		{
			if (Node->GetChild(2)->GetType() != ttFNNOOP)
				ExecRecursive(Node->GetChild(2));
		}
		// Execute next statement, if available
		if (Node->GetChild(3) != NULL)
			ExecRecursive(Node->GetChild(3));
		break;
	
	case ttFNWHILE:
		// Child 0 is the Boolean Expression
		// Child 1 is the iterative statement
		// Child 2 (if exists) is the next statement
		while (GetBoolExprVal(ExecRecursive(Node->GetChild(0))))
		{
			ExecRecursive(Node->GetChild(1));
		}
		// Execute next statement, if available
		if (Node->GetChild(2) != NULL)
			ExecRecursive(Node->GetChild(2));
		break;

	case ttFNDOUNTIL:
		// Child 0 is the iterative statement
		// Child 1 is the boolean expression
		// Child 2 is the next statement
		do {
			ExecRecursive(Node->GetChild(0));
		}
		while (GetBoolExprVal(ExecRecursive(Node->GetChild(1))));
		// Execute next statement, if available
		if (Node->GetChild(2) != NULL)
			ExecRecursive(Node->GetChild(2));
		break;

	case ttFNMAP:
		// Child 0 is the Attribute
		// Child 1 is the Value Clause(s)
		// Child 2 is the next statement
		ExecValues(AttribTable->GetAttrib(Node->GetChild(0)->GetVal()), Node->GetChild(1));

		// Execute next statement, if available
		if (Node->GetChild(2) != NULL)
			ExecRecursive(Node->GetChild(2));
		break;
		
	case ttFNPUSH:
		// Child 0 is the Attribute
		// Child 1 is the Expression
		// Child 2 is the next statement

		ExprValue = ExecRecursive(Node->GetChild(1));
		AttribTable->SetAttrib(Node->GetChild(0)->GetVal(),
							   ExprValue + CString("|") + AttribTable->GetAttrib(Node->GetChild(0)->GetVal()));
		// Execute next statement, if available
		if (Node->GetChild(2) != NULL)
			ExecRecursive(Node->GetChild(2));
		break;

	case ttFNPOP:
		// Child 0 is the source Attribute
		// Child 1 is the dest Attribute
		// Child 2 is the next statement
		AttrVal = AttribTable->GetAttrib(Node->GetChild(0)->GetVal());

		if (AttrVal.GetLength() == 0)
		{
			// Pop underflow
			return CString("");
		}
		else
		{
			if (AttrVal.Find("|") > -1)
			{
				RetVal = AttrVal.Mid(0, AttrVal.Find("|"));
				AttrVal.Delete(0, AttrVal.Find("|") + 1);
				AttribTable->SetAttrib(Node->GetChild(0)->GetVal(), AttrVal);
				AttribTable->SetAttrib(Node->GetChild(1)->GetVal(), RetVal);
			}
			else
			{
				RetVal = AttrVal;
				AttribTable->SetAttrib(Node->GetChild(0)->GetVal(), CString(""));
				AttribTable->SetAttrib(Node->GetChild(1)->GetVal(), RetVal);
			}
		}
		// Execute next statement, if available
		if (Node->GetChild(2) != NULL)
			ExecRecursive(Node->GetChild(2));
		if (RetVal.GetLength() > 0)
			return RetVal;
		break;

	case ttFNREPLACE:
		// Replace command performs the following functions:
		// If the start place is somewhere in the string, replace the
		// characters in string starting at 'start', until either
		// the total number of characters to be replaced is finished,
		// or the replace-string is exhausted.  If the replacement would
		// add characters beyond the end of the original expression, it will
		// add those characters anyway, resulting in an attribute whose
		// value is longer than it was prior to the command.  Finally,
		// if the begin value is further than the end of the attribute,
		// the replacement string will simply be appended on the end.
		// NOTE that the index is 1-based.
		// Examples:  
		// 1. &A = 'abcde', command is Replace(&A, "cat", 2, 3);
		//    Results in 'acate'
		// 2. &A = 'abcde', command is Replace(&A, "ca", 2, 3);
		//    Results in 'acade'
		// 3. &A = 'abcde', command is Replace(&a, "cat", 2, 2);
		//    Results in 'acade'
		// 4. &A = 'abcde', command is Replace(&a, "cat", 4, 3);
		//    Results in 'abccat'

		// Child 0 is the Attribute
		// Child 1 is a string expression that will be used to replace
		// Child 2 is numeric expression, the 'start' place
		// Child 3 is numeric expression, the number of characters to replace
		// Child 4 is the next statement
		{
			CString TargetStr;
			CString RepStr;
			int BeginVal;
			int LengthVal;
			int index;
			
			TargetStr = AttribTable->GetAttrib(Node->GetChild(0)->GetVal());
			RepStr = ExecRecursive(Node->GetChild(1));
			BeginVal = atoi(ExecRecursive(Node->GetChild(2))) - 1;
			LengthVal = atoi(ExecRecursive(Node->GetChild(3)));

			if (BeginVal > TargetStr.GetLength())
			{
				TargetStr += RepStr;
			}
			else if (RepStr.GetLength() > BeginVal + LengthVal)
			{
				TargetStr = TargetStr.Mid(0, BeginVal);
				TargetStr += RepStr;
			}
			else
			{
				index = 0;
				while (index < LengthVal && 
					   index < RepStr.GetLength())
				{
					if (BeginVal + index >= TargetStr.GetLength())
						TargetStr += RepStr.GetAt(index);
					else
						TargetStr.SetAt(BeginVal + index, RepStr.GetAt(index));
					// Prevent 'runaway' strings.. limit to 1K characters
					if (TargetStr.GetLength() > 1024)
						TargetStr = TargetStr.Left(1024);

					index++;
				}
			}


			AttribTable->SetAttrib(Node->GetChild(0)->GetVal(), TargetStr);

		}
		// Execute next statement, if available
		if (Node->GetChild(4) != NULL)
			ExecRecursive(Node->GetChild(4));
		break;

	case ttFNGETAT:
		// Child 0 is the Attribute
		// Child 1 is a numeric expression
		// Child 2 is also a numeric expression
		// No 'next statements' as it is a function, not a statement
		{
			CString AttribStr;
			CString RetVal;
			int BeginVal;
			int LengthVal;

			AttribStr = AttribTable->GetAttrib(Node->GetChild(0)->GetVal());
			BeginVal = atoi(ExecRecursive(Node->GetChild(1)));
			if (BeginVal < 1 || BeginVal > AttribStr.GetLength())
			{
				return CString("0");
			}
			LengthVal = atoi(ExecRecursive(Node->GetChild(2)));
			if (LengthVal + BeginVal > AttribStr.GetLength())
			{
				return CString("0");
			}
			RetVal = AttribStr.Mid(BeginVal - 1, LengthVal);
			return RetVal;
		}
		break;

	
	case ttADD:
		// Child 0 is the left value
		// Child 1 is the right value
		{
			
			double DoubleVal;
			CString RetVal;
			DoubleVal = GetNumExprVal(ExecRecursive(Node->GetChild(0))) + 
						GetNumExprVal(ExecRecursive(Node->GetChild(1)));
			RetVal.Format("%f", DoubleVal);
			return RetVal;
		}
		break;

	case ttSUBTRACT:
		// Child 0 is the left value
		// Child 1 is the right value
		{
			double DoubleVal;
			CString RetVal;
			DoubleVal = GetNumExprVal(ExecRecursive(Node->GetChild(0))) - 
						GetNumExprVal(ExecRecursive(Node->GetChild(1)));
			RetVal.Format("%f", DoubleVal);
			return RetVal;
		}
		break;

	case ttMULTIPLY:
		// Child 0 is the left value
		// Child 1 is the right value
		{
			double DoubleVal;
			CString RetVal;
			DoubleVal = GetNumExprVal(ExecRecursive(Node->GetChild(0))) * 
						GetNumExprVal(ExecRecursive(Node->GetChild(1)));
			RetVal.Format("%f", DoubleVal);
			return RetVal;
		}
		break;

	case ttDIVIDE:
		// Child 0 is the left value
		// Child 1 is the right value
		// Note this is "Protected" Divide.  Anything divided by zero will
		// automatically return a 1.
		{
			double DoubleVal;
			double LeftVal;
			double RightVal;

			CString RetVal;
			
			LeftVal = GetNumExprVal(ExecRecursive(Node->GetChild(0)));
			RightVal = GetNumExprVal(ExecRecursive(Node->GetChild(1)));
			if (RightVal == 0)
				return (CString("1.0"));

			DoubleVal = LeftVal / RightVal;
			RetVal.Format("%f", DoubleVal);
			return RetVal;
		}
		break;

	case ttMOD:
		// Child 0 is the left value
		// Child 1 is the right value
		{
			long DoubleVal;
			long LeftVal;
			long RightVal;

			CString RetVal;
			
			LeftVal = (int) (GetNumExprVal(ExecRecursive(Node->GetChild(0))));
			RightVal = (int) (GetNumExprVal(ExecRecursive(Node->GetChild(1))));
			if (RightVal == 0L)
				RightVal = 1L;
			DoubleVal = LeftVal % RightVal;
			RetVal.Format("%ld", DoubleVal);
			return RetVal;
		}
		break;

	// Single-argument operators
	case ttNEGATIVE:
		// Child 0 is the only value
		{
			CString RetVal;
			RetVal.Format("%f", -1.0 * (GetNumExprVal(ExecRecursive(Node->GetChild(0)))));
			return RetVal;
		}
		break;

	case ttNOT:
		{
			bool BoolVal;
			BoolVal = !(GetBoolExprVal(ExecRecursive(Node->GetChild(0))));
			return (BoolVal ? CString("TRUE") : CString("FALSE"));
		}
		break;

	case ttAND:
		{
			bool BoolVal;
			BoolVal = GetBoolExprVal(ExecRecursive(Node->GetChild(0))) &&
					  GetBoolExprVal(ExecRecursive(Node->GetChild(1)));
			return (BoolVal ? CString("TRUE") : CString("FALSE"));
		}
		break;

	case ttOR:
		{
			bool BoolVal;
			BoolVal = GetBoolExprVal(ExecRecursive(Node->GetChild(0))) ||
					  GetBoolExprVal(ExecRecursive(Node->GetChild(1)));
			return (BoolVal ? CString("TRUE") : CString("FALSE"));
		}
		break;


	// Constants
	case ttFLOAT:
	case ttBOOL:
	case ttSTRING:
		return Node->GetVal();
		break;

	case ttNUMFUNCT:
		// All sorts of different mathematical functions
		{
			CString RetVal;
			double GetVal;
			if (Node->GetVal() == CString("sin"))
			{
				GetVal = sin(GetNumExprVal(ExecRecursive(Node->GetChild(0))));
				RetVal.Format("%f", GetVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("cos"))
			{
				GetVal = cos(GetNumExprVal(ExecRecursive(Node->GetChild(0))));
				RetVal.Format("%f", GetVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("abs"))
			{
				GetVal = fabs(GetNumExprVal(ExecRecursive(Node->GetChild(0))));
				RetVal.Format("%f", GetVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("tan"))
			{
				GetVal = tan(GetNumExprVal(ExecRecursive(Node->GetChild(0))));
				RetVal.Format("%f", GetVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("log"))
			{
				GetVal = log(GetNumExprVal(ExecRecursive(Node->GetChild(0))));
				RetVal.Format("%f", GetVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("sqrt"))
			{
				GetVal = sqrt(GetNumExprVal(ExecRecursive(Node->GetChild(0))));
				RetVal.Format("%f", GetVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("abs"))
			{
				GetVal = fabs(GetNumExprVal(ExecRecursive(Node->GetChild(0))));
				RetVal.Format("%f", GetVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("round"))
			{
				long FinalVal;
				GetVal = GetNumExprVal(ExecRecursive(Node->GetChild(0))) + 0.5;
				RetVal.Format("%f", GetVal);
				FinalVal = atol(RetVal);
				RetVal.Format("%ld", FinalVal);
				return RetVal;
			}

			if (Node->GetVal() == CString("trunc"))
			{
				long FinalVal;
				GetVal = GetNumExprVal(ExecRecursive(Node->GetChild(0)));
				RetVal.Format("%f", GetVal);
				FinalVal = atol(RetVal);
				RetVal.Format("%ld", FinalVal);
				return RetVal;
			}
			// Error - Unknown function
			return CString("UNKNOWN FUNCTION");
		}
		break;

	case ttCOMPAREOP:
		{

			CString LeftVal;
			CString RightVal;
			CString RetVal;

			LeftVal = ExecRecursive(Node->GetChild(0));
			RightVal = ExecRecursive(Node->GetChild(1));

			if (Node->GetVal() == CString(">="))
			{
				RetVal.Format("%s", GetNumExprVal(LeftVal) >= GetNumExprVal(RightVal) ? CString("TRUE") : CString("FALSE"));
				return RetVal;
			}

			if (Node->GetVal() == CString("<="))
			{
				double TestL = GetNumExprVal(LeftVal);
				double TestR = GetNumExprVal(RightVal);
				RetVal.Format("%s", TestL <= TestR ? CString("TRUE") : CString("FALSE"));
				return RetVal;
			}

			if (Node->GetVal() == CString(">"))
			{
				RetVal.Format("%s", GetNumExprVal(LeftVal) > GetNumExprVal(RightVal) ? CString("TRUE") : CString("FALSE"));
				return RetVal;
			}

			if (Node->GetVal() == CString("<"))
			{
				RetVal.Format("%s", GetNumExprVal(LeftVal) < GetNumExprVal(RightVal) ? CString("TRUE") : CString("FALSE"));
				return RetVal;
			}

			if (Node->GetVal() == CString("="))
			{
				if (IsBoolean(LeftVal) != -1 && IsBoolean(RightVal) != -1)
				{
					RetVal.Format("%s", GetBoolExprVal(LeftVal) == GetBoolExprVal(RightVal) ? CString("TRUE") : CString("FALSE"));
				}
				else
				{
					if (IsNumber(LeftVal) && IsNumber(RightVal))
						RetVal.Format("%s", GetNumExprVal(LeftVal) == GetNumExprVal(RightVal) ? CString("TRUE") : CString("FALSE"));
					else
						RetVal.Format("%s", LeftVal == RightVal ? CString("TRUE") : CString("FALSE"));
				}
				return RetVal;
			}

			if (Node->GetVal() == CString("!=") ||
				Node->GetVal() == CString("<>"))
			{
				if (IsBoolean(LeftVal) && IsBoolean(RightVal))
				{
					RetVal.Format("%s", GetBoolExprVal(LeftVal) != GetBoolExprVal(RightVal) ? CString("TRUE") : CString("FALSE"));
				}
				else
				{
					if (IsNumber(LeftVal) && IsNumber(RightVal))
						RetVal.Format("%s", GetNumExprVal(LeftVal) != GetNumExprVal(RightVal) ? CString("TRUE") : CString("FALSE"));
					else
						RetVal.Format("%s", LeftVal != RightVal ? CString("TRUE") : CString("FALSE"));
				}	
			}
			return RetVal;
		}
		break;

	case ttUSRATTRIB:
		return AttribTable->GetAttrib(Node->GetVal());
		break;

	case ttUSRFUNCTION:
		{
			// The execution of the User-entered functions (in the User Functions tab)
			// Simulates a recursive evaluation, if necessary.  First, adds its parameters and
			// return value to the attribute table, then sets the parameters with recursive calls.
			// Then, it executes the user function.  Finally, it executes any next statement, if 
			// available.
			
			CString RetVal;
			int ThisArg;


			// Prevent run-away recursion -- will assume nothing goes further than 50 levels
			if (CurrentFunctLevel < 50)
			{
				// Add one to CurrentFunctLevel
				CurrentFunctLevel ++;

				// Add Function Name to the Attribute Table as a return value
				AttribTable->AddAttribute(rtSTRING, Node->GetVal(), CurrentFunctLevel);

				// Add Function Parameters to the Attribute Table
				for (ThisArg = 0; ThisArg < FunctionTable->GetParameterCount(Node->GetVal()); ThisArg++)
					AttribTable->AddAttribute(rtSTRING, FunctionTable->GetParmName(Node->GetVal(), ThisArg), CurrentFunctLevel);

				// Calculate and set the parameters
				for (ThisArg = 0; ThisArg < FunctionTable->GetParameterCount(Node->GetVal()); ThisArg++)
					AttribTable->SetAttrib(FunctionTable->GetParmName(Node->GetVal(), ThisArg), ExecRecursive(Node->GetChild(ThisArg)));

				// Execute Function
				ExecRecursive(FunctionTable->GetFunctionProgram(Node->GetVal())->ProgramTree);
				
				// Store Return value temporarily; delete return attribute from Attribute table
				RetVal = AttribTable->GetAttrib(Node->GetVal());

				// Delete any attributes whose level is higher than that of the ending function
				//    from the Attribute Table
				AttribTable->RemoveLevel(CurrentFunctLevel);

				// Subtract one from CurrentFunctLevel
				CurrentFunctLevel--;

				// Execute next statement, if available (Data-Returning functions will not)
				if (Node->GetChild(FunctionTable->GetParameterCount(Node->GetVal())) != NULL)
					ExecRecursive(Node->GetChild(FunctionTable->GetParameterCount(Node->GetVal())));
			}
			else
				RetVal = CString("0");

			return RetVal;

		}


	case ttGENETICFUNC:
		return ExecRecursive(RPBTree);
		break;

	}
	return CString("");
}


void CGPProgram::ExecValues(CString Attrib, class CSubTree *Node)
{
	if (Node == NULL)
		return;

	if (Attrib.Left(1) == CString ("2"))
		int atgadtg = 5;

	if (Node->GetChild(0)->GetVal() == Attrib)
	{
		ExecRecursive(Node->GetChild(1));
	}
	else
	{
		ExecValues(Attrib, Node->GetChild(2));
	}
}

bool   CGPProgram::GetBoolExprVal(CString ReturnedVal)
{
	if (ReturnedVal == CString("TRUE") ||
		ReturnedVal == CString("1") ||
		ReturnedVal == CString("T") ||
		ReturnedVal == CString("YES") ||
		ReturnedVal == CString("Y"))
		return TRUE;
	else
		return FALSE;
}

double CGPProgram::GetNumExprVal(CString ReturnedVal)
{
	if (ReturnedVal == CString("TRUE") ||
		ReturnedVal == CString("1") ||
		ReturnedVal == CString("T") ||
		ReturnedVal == CString("YES") ||
		ReturnedVal == CString("Y"))
		return 1.0;

	return atof(ReturnedVal);
}
